@@ -4,5 +4,7 @@ from lightwood.data.cleaner import cleaner
 from lightwood.data.splitter import splitter
 from lightwood.data.timeseries_transform import transform_timeseries
 from lightwood.data.timeseries_analyzer import timeseries_analyzer
+from lightwood.data.encoded_ds import EncodedDs, ConcatedEncodedDs
 
-__all__ = ['infer_types', 'statistical_analysis', 'cleaner', 'splitter', 'transform_timeseries', 'timeseries_analyzer']
+__all__ = ['infer_types', 'statistical_analysis', 'cleaner', 'splitter', 'transform_timeseries', 'timeseries_analyzer',
+           'EncodedDs', 'ConcatedEncodedDs']
