@@ -2,7 +2,7 @@ from lightwood.helpers.accuracy import to_binary, f1_score, recall_score, precis
 from lightwood.helpers.device import is_cuda_compatible, get_devices
 from lightwood.helpers.general import mase, is_none, evaluate_accuracy, evaluate_num_array_accuracy,\
     evaluate_array_accuracy, evaluate_multilabel_accuracy, evaluate_regression_accuracy, evaluate_cat_array_accuracy, \
-    bounded_evaluate_array_accuracy
+    bounded_ts_accuracy
 from lightwood.helpers.ts import get_group_matches, get_ts_groups, get_inferred_timestamps, add_tn_num_conf_bounds, \
     add_tn_cat_conf_bounds
 from lightwood.helpers.io import read_from_path_or_url
@@ -17,7 +17,7 @@ from lightwood.helpers.torch import average_vectors, concat_vectors_and_pad, Lig
 
 __all__ = ['to_binary', 'f1_score', 'recall_score', 'precision_score', 'r2_score', 'is_cuda_compatible', 'get_devices',
            'get_group_matches', 'get_ts_groups', 'mase', 'is_none', 'evaluate_accuracy', 'evaluate_num_array_accuracy',
-           'evaluate_array_accuracy', 'evaluate_cat_array_accuracy', 'bounded_evaluate_array_accuracy',
+           'evaluate_array_accuracy', 'evaluate_cat_array_accuracy', 'bounded_ts_accuracy',
            'evaluate_multilabel_accuracy', 'evaluate_regression_accuracy', 'read_from_path_or_url', 'get_nr_procs',
            'mut_method_call', 'run_mut_method', 'tokenize_text', 'analyze_sentences', 'decontracted', 'contains_alnum',
            'get_identifier_description', 'get_identifier_description_mp', 'get_pct_auto_increment',
